#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Synchronous low-level connection."""

import contextlib
import json as _json
import logging
from collections.abc import Callable, Generator
from functools import partial
from typing import Any, Literal, Optional, cast, overload

import requests
from attrs import define, field
from cattrs.dispatch import UnstructureHook
from requests import adapters
from requests.structures import CaseInsensitiveDict
from urllib3.util import Retry
from yarl import URL

from ...converter import deserialize_rest_response
from ...errors import (
    RepositoryClientError,
    RepositoryCommunicationError,
    RepositoryError,
    RepositoryServerError,
)
from ...progress import DummyProgressBar, ProgressBar
from ...types.auth import BearerTokenForHost
from ..streams.base import DataSink, DataSource
from ..streams.progress import ProgressSink
from .auth import BearerAuthentication
from .aws_limits import MINIMAL_DOWNLOAD_PART_SIZE, adjust_download_multipart_params
from .limiter import current_limiter

log = logging.getLogger("invenio_nrp.sync_client.connection")
communication_log = logging.getLogger("invenio_nrp.communication")


@contextlib.contextmanager
def _cast_error() -> Generator[None, None, None]:
    """Catch all errors and cast them to Repository*Error.

    :return:
    """
    try:
        yield
    except RepositoryError:
        raise
    except Exception as e:
        raise RepositoryCommunicationError(str(e)) from e


class try_until_success:
    def __init__(self, attempts: int):
        self.attempts = attempts
        self.attempt = 0
        self.done = False
        self.failures: list[Exception] = []

    def __iter__(self):
        while not self.done and self.attempt < self.attempts:
            i = self.attempt
            yield self
            assert i != self.attempt, "attempt not attempted"

        if self.done:
            return

        if self.failures:
            raise ExceptionGroup("Failures in HTTP transport", self.failures)

    def __enter__(self):
        self.attempt += 1

    def __exit__(self, _ext, exc, _tb):  # type: ignore
        if exc:
            if isinstance(exc, RepositoryClientError):
                return False
            else:
                self.failures.append(exc)  # type: ignore
                return True

        self.done = True


def raise_for_invenio_status(response: requests.Response) -> None:
    """Raise an exception if the response is not successful."""
    payload: dict[str, Any]
    if not response.ok:
        payload_text = response.text
        try:
            payload = _json.loads(payload_text)
        except ValueError:
            payload = {
                "status": response.status_code,
                "reason": payload_text,
            }

        if response.status_code >= 500:
            raise RepositoryServerError(response.request, payload)
        elif response.status_code >= 400:
            raise RepositoryClientError(response.request, payload)
        raise RepositoryCommunicationError(response.request, payload)


class SyncConnection:
    """Low-level synchronous connection to the repository."""

    def __init__(
        self,
        tokens: dict[URL, str] | None = None,
        verify_tls: bool = True,
        retry_count: int = 5,
        retry_after_seconds: int = 1,
    ):
        """Initialize the connection."""
        self._verify_tls = verify_tls
        self._retry_count = retry_count
        self._retry_after_seconds = retry_after_seconds

        _tokens: list[BearerTokenForHost] = [
            BearerTokenForHost(host_url=url, token=token)
            for url, token in (tokens or {}).items()
            if token
        ]
        self._auth = BearerAuthentication(_tokens)

    @contextlib.contextmanager
    def _client(
        self, idempotent: bool = False
    ) -> Generator[requests.Session, None, None]:
        """Create a new session with the repository and configure it with the token.

        :return: A new http client
        """
        """
        Create a new session with the repository and configure it with the token.
        :return: A new http client
        """
        if idempotent:
            retry = Retry(
                total=self._retry_count,
                backoff_factor=self._retry_after_seconds,
                status_forcelist=[429, 500, 502, 503, 504],
                respect_retry_after_header=True,
            )
            adapter = adapters.HTTPAdapter(max_retries=retry)
        else:
            adapter = adapters.HTTPAdapter()

        session = requests.Session()
        session.auth = self._auth
        session.mount("https://", adapter)
        session.verify = self._verify_tls

        yield session

    @overload
    def head(
        self,
        *,
        url: URL,
        use_get: bool = False,
        get_links: Literal[False],
        **kwargs: Any,  # noqa: ANN401
    ) -> CaseInsensitiveDict[str]: ...

    @overload
    def head(
        self,
        *,
        url: URL,
        use_get: bool = False,
        get_links: Literal[True],
        **kwargs: Any,  # noqa: ANN401
    ) -> dict[str, URL]: ...

    def head(
        self,
        *,
        url: URL,
        use_get: bool = False,
        get_links: bool = False,
        **kwargs: Any,  # noqa: ANN401
    ) -> CaseInsensitiveDict[str] | dict[str, URL]:
        """Perform a HEAD request to the repository.

        :param url:                 the url of the request
        :param idempotent:          True if the request is idempotent, should be for HEAD requests
        :param kwargs:              any kwargs to pass to the aiohttp client
        :return:                    None

        :raises RepositoryClientError: if the request fails due to client passing incorrect parameters (HTTP 4xx)
        :raises RepositoryServerError: if the request fails due to server error (HTTP 5xx)
        :raises RepositoryCommunicationError: if the request fails due to network
        """

        def _head(
            response: requests.Response,
        ) -> CaseInsensitiveDict[str] | dict[str, URL]:
            if get_links:
                return {k: URL(v) for k, v in response.links.items()}
            else:
                return response.headers

        if use_get:
            return self._retried(
                "GET", url, _head, idempotent=True, headers={"Range": "bytes=0-0"}
            )
        else:
            return self._retried("HEAD", url, _head, idempotent=True)

    def get[T](
        self,
        *,
        url: URL,
        idempotent: bool = True,
        result_class: type[T],
        **kwargs: Any,  # noqa: ANN401
    ) -> T:
        """Perform a GET request to the repository.

        :param url:
        :param idempotent:
        :param result_class:
        :param kwargs:
        :return:

        :raises RepositoryClientError: if the request fails due to client passing incorrect parameters (HTTP 4xx)
        :raises RepositoryServerError: if the request fails due to server error (HTTP 5xx)
        :raises RepositoryCommunicationError: if the request fails due to network error
        """
        return self._retried(
            "GET",
            url,
            partial(
                self._get_call_result,
                result_class=result_class,
            ),
            idempotent=True,
            **kwargs,
        )

    def post[T](
        self,
        *,
        url: URL,
        json: dict[str, Any] | list[Any] | None = None,
        data: bytes | None = None,
        idempotent: bool = False,
        result_class: type[T] | None,
        **kwargs: Any,  # noqa: ANN401
    ) -> T | None:
        """Perform a POST request to the repository.

        :param url:
        :param json:
        :param data:
        :param idempotent:
        :param result_class:
        :param kwargs:
        :return:

        :raises RepositoryClientError: if the request fails due to client passing incorrect parameters (HTTP 4xx)
        :raises RepositoryServerError: if the request fails due to server error (HTTP 5xx)
        :raises RepositoryCommunicationError: if the request fails due to network
        """
        assert (
            json is not None or data is not None
        ), "Either json or data must be provided"

        if result_class:
            return self._retried(
                "POST",
                url,
                partial(
                    self._get_call_result,
                    result_class=result_class,
                ),
                idempotent=idempotent,
                json=json,
                data=data,
                **kwargs,
            )
        else:
            return self._retried(
                "POST",
                url,
                self._check_empty_call_result,
                idempotent=idempotent,
                json=json,
                data=data,
                **kwargs,
            )

    def put[T](
        self,
        *,
        url: URL,
        json: dict[str, Any] | list[Any] | None = None,
        data: bytes | None = None,
        result_class: type[T],
        **kwargs: Any,  # noqa: ANN401
    ) -> T:
        """Perform a PUT request to the repository.

        :param url:
        :param json:
        :param data:
        :param result_class:
        :param kwargs:
        :return:

        :raises RepositoryClientError: if the request fails due to client passing incorrect parameters (HTTP 4xx)
        :raises RepositoryServerError: if the request fails due to server error (HTTP 5xx)
        :raises RepositoryCommunicationError: if the request fails due to network
        """
        assert (
            json is not None or data is not None
        ), "Either json or data must be provided"

        return self._retried(
            "PUT",
            url,
            partial(
                self._get_call_result,
                result_class=result_class,
            ),
            idempotent=True,
            json=json,
            data=data,
            **kwargs,
        )

    def put_stream(
        self,
        *,
        url: URL,
        source: DataSource,
        open_kwargs: dict[str, Any] | None = None,
        **kwargs: Any,  # noqa: ANN401
    ) -> requests.Response:
        """Perform a PUT request to the repository with a file.

        :param url:
        :param file:
        :param kwargs:
        :return:

        :raises RepositoryClientError: if the request fails due to client passing incorrect parameters (HTTP 4xx)
        :raises RepositoryServerError: if the request fails due to server error (HTTP 5xx)
        :raises RepositoryCommunicationError: if the request fails due to network
        """

        def _put(response: requests.Response) -> requests.Response:
            raise_for_invenio_status(response)
            return response

        return self._retried(
            "PUT",
            url,
            _put,
            idempotent=True,
            data=partial(source.open, **(open_kwargs or {})),
            **kwargs,
        )

    def get_stream(
        self,
        *,
        url: URL,
        sink: DataSink,
        offset: int = 0,
        size: int | None = None,
        **kwargs: Any,  # noqa: ANN401
    ) -> None:
        """Perform a GET request to the repository and write the response to a sink.

        :param url:                 the url of the request
        :param kwargs:              any kwargs to pass to the aiohttp client
        :return:                    the parsed result

        :raises RepositoryClientError: if the request fails due to client passing incorrect parameters (HTTP 4xx)
        :raises RepositoryServerError: if the request fails due to server error (HTTP 5xx)
        :raises RepositoryCommunicationError: if the request fails due to network error
        """

        def _copy_stream(response: requests.Response) -> None:
            chunk = sink.open_chunk(offset=offset)
            try:
                for data in response.iter_content(chunk_size=16384):
                    chunk.write(data)
            finally:
                chunk.close()

        if size is not None:
            range_header = f"bytes={offset}-{offset + size - 1}"
        else:
            range_header = f"bytes={offset}-"

        return self._retried(
            "GET",
            url,
            _copy_stream,
            idempotent=True,
            headers={"Range": range_header},
            stream=True,
            **kwargs,
        )

    def delete(
        self,
        *,
        url: URL,
        idempotent: bool = False,
        **kwargs: Any,  # noqa: ANN401
    ) -> None:
        """Perform a DELETE request to the repository.

        :param url:
        :param idempotent:
        :param kwargs:
        :return:
        :raises RepositoryClientError: if the request fails due to client passing incorrect parameters (HTTP 4xx)
        :raises RepositoryServerError: if the request fails due to server error (HTTP 5xx)
        :raises RepositoryCommunicationError: if the request fails due to network
        """
        return self._retried(
            "DELETE",
            url,
            self._check_empty_call_result,
            idempotent=idempotent,
            **kwargs,
        )

    def _check_empty_call_result(self, response: requests.Response) -> None:
        raise_for_invenio_status(response)
        if response.status_code == 204:
            return None  # No content
        raise RepositoryCommunicationError("Expected empty response")

    def _get_call_result[T](
        self,
        response: requests.Response,
        result_class: type[T],
    ) -> T:
        """Get the result from the response.

        :param response:
        :param result_class:
        :return:

        :raises RepositoryClientError: if the request fails due to client passing incorrect parameters (HTTP 4xx)
        :raises RepositoryServerError: if the request fails due to server error (HTTP 5xx)
        :raises RepositoryCommunicationError: if the request fails due to network
        """
        raise_for_invenio_status(response)
        if response.status_code == 204:
            raise RepositoryCommunicationError("Expected response with content")

        json_payload = response.content
        if communication_log.isEnabledFor(logging.INFO):
            communication_log.info(
                "%s", _json.dumps(_json.loads(json_payload))
            )
        if issubclass(result_class, requests.Response):
            return cast("T", response)    # mypy can not get it
        elif issubclass(result_class, str):
            return cast("T", json_payload.decode('utf-8'))    # mypy can not get it
        elif issubclass(result_class, dict):
            return _json.loads(json_payload)
        etag = remove_quotes(response.headers.get("ETag"))
        return deserialize_rest_response(
            self, json_payload, result_class, etag
        )

    def _retried[T](
        self,
        method: str,
        url: URL,
        callback: Callable[[requests.Response], T] | None,
        idempotent: bool,
        **kwargs: Any,  # noqa: ANN401
    ) -> T:
        """Log the start of a request and retry it if necessary."""
        json = kwargs.get("json")
        if json is not None and callable(json):
            json = json()
            kwargs["json"] = json

        data = kwargs.get("data")

        if communication_log.isEnabledFor(logging.INFO):
            communication_log.info("%s %s", method.upper(), url)
            if json is not None:
                communication_log.info("%s", _json.dumps(json))
            if data is not None:
                communication_log.info("(stream)")

        for attempt in try_until_success(self._retry_count if idempotent else 1):
            actual_data = None
            if data is not None and callable(data):
                actual_data = data()
                kwargs["data"] = actual_data

            try:
                with (
                    attempt,
                    self._client(idempotent=True) as client,
                    current_limiter.limit(url),
                    _cast_error(),
                    client.request(
                        method, str(url), auth=self._auth, **kwargs
                    ) as response,
                ):
                    raise_for_invenio_status(response)  # type: ignore
                    if callback is not None:
                        return callback(response)
                    return None
            finally:
                if actual_data is not None and hasattr(actual_data, "close"):
                    actual_data.close()

        raise Exception("unreachable")

    def download_file(
        self,
        url: URL,
        sink: DataSink,
        parts: int | None = None,
        part_size: int | None = None,
        progress_bar: ProgressBar | None = None,
    ) -> None:
        progress_bar = progress_bar or DummyProgressBar()
        try:
            headers = self.head(url=url, get_links=False)
        except RepositoryClientError:
            # The file is not available for HEAD. This is the case for S3 files
            # where the file is a pre-signed request. We'll try to download the headers
            # with a GET request with a range header containing only the first byte.
            headers = self.head(url=url, use_get=True, get_links=False)

        size = 0
        location = URL(headers.get("Location", url))

        if "Content-Range" in headers:
            size = int(headers["Content-Range"].split("/")[-1])
        elif "Content-Length" in headers:
            size = int(headers["Content-Length"])

        if size:
            sink.allocate(size)
            progress_bar.set_total(size)

        if (
            size
            and size > MINIMAL_DOWNLOAD_PART_SIZE
            and headers.get("Accept-Ranges") == "bytes"
        ):
            self._download_multipart(location, sink, size, progress_bar, parts, part_size)
        else:
            self._download_single(location, sink, progress_bar)

    def _download_single(self, url: URL, sink: DataSink, progress_bar: ProgressBar) -> None:
        self.get_stream(url=url, sink=ProgressSink(sink, progress_bar), offset=0)

    def _download_multipart(
        self,
        url: URL,
        sink: DataSink,
        size: int,
        progress_bar: ProgressBar,
        parts: int | None = None,
        part_size: int | None = None,
    ) -> None:
        adjusted_part_size, adjusted_parts = adjust_download_multipart_params(
            size, parts, part_size
        )

        for i in range(adjusted_parts):
            start = i * adjusted_part_size
            part_size = min((i + 1) * adjusted_part_size, size) - start
            self.get_stream(url=url, sink=ProgressSink(sink, progress_bar), offset=start, size=part_size)


def remove_quotes(etag: str | None) -> Optional[str]:
    if etag is None:
        return None
    return etag.strip('"')


def connection_unstructure_hook(data: Any, previous: UnstructureHook) -> Any:
    ret = previous(data)
    ret.pop("_connection", None)
    ret.pop("_etag", None)
    return ret


@define(kw_only=True)
class ConnectionMixin:
    """A mixin for classes that are a result of a REST API call."""

    _connection: SyncConnection = field(init=False, default=None)
    """Connection is automatically injected"""

    _etag: Optional[str] = field(init=False, default=None)
    """etag is automatically injected if it was returned by the repository"""

    def _set_connection_params(
        self, connection: SyncConnection, etag: Optional[str] = None
    ) -> None:
        """Set the connection and etag."""
        self._connection = connection
        self._etag = etag

    def _etag_headers(self) -> dict[str, str]:
        """Return the headers with the etag if it was returned by the repository."""
        headers: dict[str, str] = {}
        if self._etag:
            headers["If-Match"] = self._etag
        return headers


__all__ = ("SyncConnection",)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Transfer protocols."""
from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from yarl import URL

    from ....progress import ProgressBar
    from ...connection import AsyncConnection
    from ..files import File
    from ..source import DataSource


class Transfer(Protocol):
    """Protocol for transferring files to the repository."""

    async def prepare(
        self,
        connection: AsyncConnection,
        files_link: URL,
        transfer_payload: dict,
        source: DataSource,
    ) -> None:
        """Prepare the transfer.

        :param connection:              connection to the repository
        :param files_link:              link where the files are to be uploaded
        :param transfer_payload:        extra payload for the transfer, might be freely modified by this method
        """
        ...

    async def upload(
        self,
        connection: AsyncConnection,
        initialized_upload: File,
        source: DataSource,
        progress_bar: ProgressBar,
    ) -> None:
        """Upload the file.

        :param connection:              connection to the repository
        :param initialized_upload:      initialized upload as returned from the repository
        :param file:                    file to be uploaded
        """
        ...

    async def get_commit_payload(self, initialized_upload: File) -> dict:
        """Finalize the successful upload.

        There is no method to discard an unsuccessful upload,
        just use the "delete" operation on the failed file.

        :param initialized_upload:      initialized upload as returned from the repository
        :return:                        metadata of the upload as returned from the repository
        """
        ...



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Commandline client for updating metadata of files."""

from typing import Optional

from rich.console import Console

from nrp_cmd.async_client import limit_connections
from nrp_cmd.cli.base import async_command
from nrp_cmd.cli.records.get import read_record
from nrp_cmd.config import Config

from ..arguments import (
    Model,
    Output,
    VerboseLevel,
    argument_with_help,
    with_config,
    with_model,
    with_output,
    with_repository,
    with_resolved_vars,
)


@with_config
@with_repository
@with_resolved_vars("record_id")
@with_output
@with_model
@argument_with_help("record_id", type=str, help="Record ID")
@argument_with_help("key", type=str, help="Key for the file")
@async_command
async def delete_file(
    *,
    config: Config,
    repository: Optional[str] = None,
    out: Output,
    record_id: str,
    key: str,
    model: Model,
) -> None:
    """Delete a file in a record."""
    console = Console()

    with limit_connections(10):
        (
            record,
            record_id_url,
            _repository_config,
            _record_client,
            repository_client,
        ) = await read_record(
            record_id,
            repository,
            config,
            False,
            model.model,
            model.published,
            model.draft,
        )

        files_client = repository_client.files
        await files_client.delete(record, key)

    if out.verbosity != VerboseLevel.QUIET:
        console.print(f"[green]Deleted file {key} in record {record_id_url}.[/green]")



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Commandline client for uploading files."""

from asyncio import Task, TaskGroup
from functools import partial
from pathlib import Path
from typing import Any, Optional

import rich_click as click
from rich.console import Console

from nrp_cmd.async_client import AsyncRepositoryClient, limit_connections
from nrp_cmd.async_client.streams import DataSource, StdInDataSource
from nrp_cmd.async_client.streams.file import FileSource
from nrp_cmd.cli.base import OutputWriter, async_command
from nrp_cmd.cli.files.table_formatters import format_files_table
from nrp_cmd.cli.records.get import read_record
from nrp_cmd.cli.records.metadata import read_metadata
from nrp_cmd.cli.records.record_file_name import create_output_file_name
from nrp_cmd.config import Config
from nrp_cmd.progress import show_progress
from nrp_cmd.types.files import File
from nrp_cmd.types.records import Record

from ..arguments import (
    Model,
    Output,
    argument_with_help,
    with_config,
    with_model,
    with_output,
    with_progress,
    with_repository,
    with_resolved_vars,
    with_verbosity,
)


async def upload_files_to_record(
    client: AsyncRepositoryClient,
    record: Record,
    *files: tuple[str | DataSource | Path, dict[str, Any] | str],
    transfer_type: str = "L",
) -> list[File]:
    """Upload files to a record."""
    # convert files to pairs
    file_client = client.files

    tasks: list[Task[Any]] = []
    async with TaskGroup() as tg:
        for _file, metadata in files:
            metadata_json: dict[str, Any]
            if not isinstance(metadata, dict):
                metadata_json = read_metadata(metadata)
            else:
                metadata_json = metadata

            key = metadata_json.get("key")
            if _file == "-":
                _file = StdInDataSource()
                key = key or "stdin"
            elif not key and isinstance(_file, (str, Path)):
                key = Path(_file).name
            if not key:
                raise ValueError("Key must be provided for file")

            if transfer_type == "M":
                # only use multipart for larger files
                if isinstance(_file, DataSource):
                    fs = await _file.size()
                elif isinstance(_file, (str, Path)):  # type: ignore
                    fs = await FileSource(_file).size()
                else:
                    raise ValueError("Invalid file source")

                if fs < 10_000_000:
                    transfer_type = "L"

            tasks.append(
                tg.create_task(
                    file_client.upload(
                        record,
                        key,
                        metadata_json,
                        _file,
                        transfer_type=transfer_type,
                        progress=key if True else None,
                    )
                )
            )
    return [t.result() for t in tasks]


@with_config
@with_repository
@with_resolved_vars("record_id")
@with_output
@with_verbosity
@with_model
@with_progress
@argument_with_help("record_id", type=str, help="Record ID")
@argument_with_help("file", type=str, help="File to upload")
@argument_with_help("metadata", type=str, required=False, help="Metadata for the file")
@click.option("--key", type=str, help="Key for the file")
@click.option("--transfer-type", type=str, help="Transfer type")
@async_command
async def upload_files(
    *,
    config: Config,
    repository: Optional[str] = None,
    record_id: str,
    file: str,
    metadata: Optional[str] = None,
    key: Optional[str] = None,
    model: Model,
    transfer_type: Optional[str] = None,
    out: Output,
) -> None:
    """Upload a file to a record."""
    console = Console()
    with limit_connections(10):
        (
            record,
            _record_id,
            repository_config,
            _record_client,
            repository_client,
        ) = await read_record(
            record_id,
            repository,
            config,
            False,
            model.model,
            model.published,
            model.draft,
        )

        metadata = metadata or "{}"
        metadata_json = read_metadata(metadata)
        assert isinstance(metadata_json, dict), "Metadata must be a dictionary."
        if key:
            metadata_json["key"] = key
        with show_progress(total=1, quiet=not out.progress):
            if not transfer_type:
                assert repository_config.info, (
                    "Do not have info for this repository to get transfer type, "
                    "please specify it manually."
                )

                transfer_type = "M" if "M" in repository_config.info.transfers else "L"

            files = await upload_files_to_record(
                repository_client,
                record,
                (file, metadata_json),
                transfer_type=transfer_type,
            )

    if out.output:
        output = create_output_file_name(
            out.output, str(record.id), record, out.output_format
        )
        if output.parent:
            output.parent.mkdir(parents=True, exist_ok=True)
    else:
        output = None

    with OutputWriter(
        output,
        out.output_format,
        console,
        partial(format_files_table, record, verbosity=out.verbosity),
    ) as printer:
        printer.output(files)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Table formatters for files."""

from collections.abc import Generator
from typing import Any

from rich import box
from rich.table import Table

from nrp_cmd.converter import converter
from nrp_cmd.types.files import File
from nrp_cmd.types.records import Record

from ..arguments import VerboseLevel


def format_files_table(
    record: Record, data: list[File], verbosity: VerboseLevel, **kwargs: Any
) -> Generator[Table, None, None] | Generator[str, None, None]:
    """Format the files table."""
    if verbosity == VerboseLevel.QUIET:
        yield "\n".join(sorted([file.key for file in data]))
    elif verbosity == VerboseLevel.VERBOSE:
        for file in sorted(data, key=lambda f: f.key):
            table = Table(f"File {file.key} @ record {record.links.self}", box=box.SIMPLE, title_justify="left")
            table.add_row("Key", file.key)
            table.add_row("Status", file.status)
            table.add_row("Mimetype", file.mimetype)
            table.add_row("Access", str(file.access))
            table.add_row("Metadata", str(file.metadata))
            table.add_row("Size", str(file.size))
            for link_name, link in converter.unstructure(file.links).items():
                table.add_row(link_name, str(link))
            yield table
    else:
        table = Table("Key", "Size", "Content", box=box.SIMPLE, title_justify="left")
        for file in sorted(data, key=lambda f: f.key):
            table.add_row(file.key, str(file.size), str(file.links.content))
        yield table



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Errors raised from invenio repository."""

import json
from typing import Any


class RepositoryError(Exception):
    """Base class for all repository errors."""

    pass


class RepositoryCommunicationError(RepositoryError):
    """Base class for all repository communication errors."""


class RepositoryNetworkError(RepositoryCommunicationError):
    """Raised when a network error occurs."""


class RepositoryJSONError(RepositoryCommunicationError):
    """Raised from a repository when an error occurs."""

    def __init__(self, request_info: Any, response: dict[str, Any]):  # noqa ANN401
        """Initialize the error."""
        self._request_info = request_info
        self._response = response

    @property
    def request_info(self) -> Any:
        """Return the request info."""
        return self._request_info

    @property
    def json(self) -> dict[str, Any]:
        """Return the JSON response."""
        return self._response

    def __repr__(self) -> str:
        """Return the representation of the error."""
        return f"{self._request_info.url} : {json.dumps(self.json)}"

    def __str__(self) -> str:
        """Return the string representation of the error."""
        return self.__repr__()


class RepositoryServerError(RepositoryJSONError):
    """An error occurred on the server side (5xx http status code)."""


class RepositoryClientError(RepositoryJSONError):
    """An error occurred on the client side (4xx http status code).

    This is usually not found, unauthorized, malformed request and similar.
    """

class DoesNotExistError(RepositoryClientError):
    pass

class RepositoryRetryError(Exception):
    def __init__(self, after_seconds: float | None = None):
        self.after_seconds = after_seconds


class UnstructureError(Exception):
    pass


class StructureError(Exception):
    pass


def is_instance_of_exceptions(
    exception: Any, exceptions: tuple[type[Exception], ...] | type[Exception]
) -> bool:  
    # Check if the current exception is an instance of T
    if isinstance(exception, exceptions):
        return True

    # Handle ExceptionGroup (Python 3.11+)
    if isinstance(exception, BaseExceptionGroup):
        sub_exception: Any
        for sub_exception in exception.exceptions:  # type: ignore
            if is_instance_of_exceptions(sub_exception, exceptions):
                return True

    # Check chained exceptions (__cause__ or __context__)
    if exception.__cause__ is not None and is_instance_of_exceptions(
        exception.__cause__, exceptions
    ):
        return True
    if exception.__context__ is not None and is_instance_of_exceptions(
        exception.__context__, exceptions
    ):
        return True

    return False



"""
# NRP Commandline Tools and Python Client

## Overview

This package provides a set of libraries and command-line tools for interacting 
with repositories that conform to the Czech National Repository Platform. 
Although support currently focuses on InvenioRDM-based repositories, the 
architecture is designed to be easily extensible to other repository types.

## Configuration

The package uses a configuration file stored in the user's home directory under
the `.nrp` folder. You can add a new repository to the configuration by running:

```bash
nrp-cmd add repository https://my-repository.org [--alias my-repo]
```

This command will guide you through setting up the repository URL and any 
required authentication tokens. If you prefer not to use the built-in 
configuration mechanism, you can provide the necessary parameters directly 
to the client.

## API

This package offers both synchronous and asynchronous clients for working 
with the configured repositories. Choose the asynchronous client if you have 
an asyncio-based application or need higher performance for data transfers. 
For simpler applications, the synchronous client is often sufficient.

Both clients share the same high-level API, allowing you to switch between them 
easily (with the appropriate `async`/`await` adjustments as needed).

## Example

A simple script that searches for records containing the word "Einstein" in metadata
and downloads all matching records to a local directory. By default it will download
data in 10 concurrent connections (see :func:`nrp_cmd.get_async_client` for more details).

```python
from nrp_cmd import get_async_client
from nrp_cmd.async_client import AsyncRepositoryClient, download

async def run():
    client: AsyncRepositoryClient = await get_async_client("my-repo")
    client: AsyncRepositoryClient = await get_async_client("https://my-repository.org")

    async for record in client.records.scan(q="Einstein"):
        print(record.metadata)
        # and store the record together with files in a directory for further processing
        download(record, f"/path/to/download/{record.id}", with_files=True)
    
if __name__ == "__main__":
    import asyncio
    asyncio.run(run())
```

See :class:`nrp_cmd.async_client.base_client.AsyncRepositoryClient` for more details.

"""

from . import async_client, config, types
from .async_client import get_async_client

__all__ = ("get_async_client", "async_client", "config", "types")



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Formatter for request and request types tables."""

from collections.abc import Generator
from typing import Any

from rich import box
from rich.table import Table

from nrp_cmd.cli.base import write_table_row
from nrp_cmd.converter import converter
from nrp_cmd.types.requests import Request, RequestType


def format_request_table(data: Request, **kwargs: Any) -> Generator[Table, None, None]:
    """Format request table both for requests and request types."""
    table = Table(
        title=f"Request {data.id}", box=box.SIMPLE, title_justify="left"
    )
    write_table_row(table, "status", data.status)
    write_table_row(table, "type", data.type)
    write_table_row(table, "created", data.created)
    write_table_row(table, "updated", data.updated)
    write_table_row(table, "topic", data.topic)
    write_table_row(table, "receiver", data.receiver)
    write_table_row(table, "sender", data.created_by)
    if data.payload:
        if data.payload.draft_record:
            write_table_row(table, "draft record", data.payload.draft_record.links.self_)
            write_table_row(table, "draft record html", data.payload.draft_record.links.self_html)
        if data.payload.published_record:
            write_table_row(table, "published record", data.payload.published_record.links.self_)
            write_table_row(table, "published record html", data.payload.published_record.links.self_html)
        if len(data.payload._extra_data or {}) > 0:
            write_table_row(table, "payload", data.payload._extra_data)
    write_table_row(table, "links", "")
    write_table_row(table, "    self", data.links.self_)
    write_table_row(table, "    self_html", data.links.self_html)
    write_table_row(table, "    comments", data.links.comments)
    write_table_row(table, "    timeline", data.links.timeline)
    write_table_row(table, "actions", "")
    for action_name, action_link in converter.unstructure(data.links.actions).items():
        write_table_row(table, f"    {action_name}", action_link)
    yield table


def format_request_type_table(data: RequestType) -> Generator[Table, None, None]:
    """Format request type table."""
    table = Table(
        title="Request Type",
        box=box.SIMPLE,
        title_justify="left",
        show_header=False,
    )
    write_table_row(table, "id", data.type_id)
    write_table_row(table, "actions", "")
    for action_name, action_link in converter.unstructure(data.links.actions).items():
        write_table_row(table, f"    {action_name}", action_link)
    yield table


def format_request_and_types_table(
    data: dict[str, Any], **kwargs: Any
) -> Generator[Table, None, None]:
    """Format request and request types table."""
    request: Request
    for request in converter.structure(data["requests"], list[Request]):
        yield from format_request_table(request)

    request_type: RequestType
    for request_type in converter.structure(data["request_types"], list[RequestType]):
        yield from format_request_type_table(request_type)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Data source that reads data from standard input."""

from pathlib import Path

from .base import DataSource, InputStream
from .os import open_file


class StdInDataSource(DataSource):
    """A data source that reads data from standard input."""

    def __init__(self) -> None:
        super().__init__()
        self._opened = False

    async def open(self, offset: int = 0, count: int | None = None) -> InputStream:
        """Open the data source for reading."""
        if self._opened:
            raise RuntimeError("Cannot open the same data source multiple times.")
        self._opened = True
        if count is not None:
            raise ValueError("Cannot read a bounded stream from standard input.")
        if offset != 0:
            raise ValueError("Cannot seek in standard input.")
        ret = await open_file(Path("/sys/stdin"), mode="rb")
        return ret

    async def size(self) -> int:
        """Return the size of the data - in this case -1 as unknown."""
        return -1

    async def content_type(self) -> str:
        """Return the content type of the data."""
        return "application/octet-stream"

    async def close(self) -> None:
        """Close the data source."""
        pass

    async def checksum(
        self, algo: str = "md5", offset: int = 0, count: int | None = None
    ) -> str:
        raise NotImplementedError("Checksums are not supported for standard input.")

    def supported_checksums(self) -> list[str]:
        """Return a list of supported checksum algorithms."""
        return []
